package me.msoucy.ptures.model

import me.msoucy.ptures.view.BattleView

class Engine(private vararg val creatures : Pair<Creature, Int>) {
    init {
        assert(creatures.isNotEmpty())
    }

    private var currentCreature = 0

    val attacker : Creature get() {
        return creatures[currentCreature].first
    }

    val forcedSkills = mutableMapOf<Creature, Skill>()

    val activeCreatures get() = creatures.withIndex()
            .filter { (_, c) -> !c.first.hasStatus<KnockedOut>() }
            .sortedBy { (_, c) -> c.first.spd }
            .map { (i, _) -> i}

    fun resolveTurn(view : BattleView) {
        forcedSkills.clear()
        // All preconditions
        for (i in activeCreatures) {
            val creature = creatures[i]
            for (status in creature.first.statuses) {
                status.onTurnStart(this, creature.first)
            }
        }
        // Get the moves each creature will use this turn
        val moves = activeCreatures.map { i ->
            if (creatures[i].first in forcedSkills) {
                Pair(forcedSkills[creatures[i].first], nextOpponent(i))
            } else {
                TODO("Fill in view here")
            }
        }
        // Resolve each move
        for (i in activeCreatures) {
            // Resolve move
            val (skill, target) = moves[i]
            currentCreature = i
            for (step in skill.damageSteps) {
                for (t in getTargetList(step.target, target)) {
                    val (targetCreature, _) = creatures[t]
                    if (attacker.hits(targetCreature, step)) {
                        targetCreature.apply(step, attacker)
                        step.applyStatus(targetCreature)
                    }
                }
            }
            for(step in skill.postSteps) {
                for (t in getTargetList(step.target, target)) {
                    val (targetCreature, _) = creatures[t]
                    step.apply(targetCreature)
                }
            }
        }
        // All post conditions
        for (i in activeCreatures) {
            val creature = creatures[i]
            for (status in creature.first.statuses) {
                status.onTurnEnd(this)
            }
        }
    }

    private fun getTargetList(target : Target, selected : Int) : List<Int> {
        return when(target) {
            Target.Self -> listOf(currentCreature)
            Target.Selected -> if (selected == -1) { listOf() } else { listOf(selected) }
            Target.Others -> creatures.indices.filter { it != currentCreature }
            Target.Opponents -> creatures.indices.filter { creatures[it].second != creatures[currentCreature].second }
            Target.All -> creatures.indices.toList()
        }
    }

    private fun nextOpponent(idx : Int) : Int {
        var nextIdx = idx + 1
        while (nextIdx != idx) {
            if (creatures[nextIdx].second != creatures[idx].second) {
                return nextIdx
            }
            nextIdx = (nextIdx + 1) % creatures.size
        }
        // There are no opponents... so use the "none" index
        return -1
    }
}